#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <profile.h>

//#define USE_CLOCK_GET_TIME
#define USE_GET_TIME_OF_DAY

static int profile_idx;

#define PROFILE_NUM_EVENTS 32

static struct {
	const char *file;
	const char *func;
	const char *call_str;
	int line;
#ifdef USE_GET_TIME_OF_DAY
	struct timeval tv;
#endif
#ifdef USE_CLOCK_GET_TIME
	struct timespec ts;
#endif
	int us_last;
	int us_max;
	int us_min;
	long long us_total;
	int num_hits;
	int in_progress;
	int new_data_to_print;
} profile_event[PROFILE_NUM_EVENTS];

#define MAX(a,b) ((a) > (b) ? (a) : (b))

void profile_print_stats(int show_only_new)
{
	int i;
	printf("================================================================\n");
	printf("%3s%8s%8s%8s%8s%8s%8s  %s\n",
			"tp#",
			"hits#",
			"last.us",
			"min.us",
			"max.us",
			"avg.us",
			"var.\%",
			"location");
	printf("================================================================\n");
	for (i = 0; i < profile_idx; i++) {
		if ((!show_only_new && profile_event[i].num_hits) ||
				profile_event[i].new_data_to_print) {
			/* calculate missing stats */
			int us_avg =
			    profile_event[i].us_total /
			    profile_event[i].num_hits;
			int var = us_avg ?  (profile_event[i].us_max - 
				 profile_event[i].us_min) * 100 / us_avg : -1;
			printf("%3d%8d%8d%8d%8d%8d%8d  id=\"%s\" at %s:%d in %s()\n", 
					i,
					profile_event[i].num_hits,
					profile_event[i].us_last, 
					profile_event[i].us_min,
					profile_event[i].us_max, 
					us_avg,
					var,
					profile_event[i].call_str,
					profile_event[i].file,
					profile_event[i].line, 
					profile_event[i].func);
			profile_event[i].new_data_to_print = 0;
		}
	}
	fflush(stdout);
}

int profile_register_event(const char *file, const char *func, int line,
			   const char *call_str)
{
	if (profile_idx == PROFILE_NUM_EVENTS - 1) {
		printf("pref: too many events registered!");
		return -1;
	}
	profile_event[profile_idx].file = file;
	profile_event[profile_idx].func = func;
	profile_event[profile_idx].line = line;
	profile_event[profile_idx].call_str = call_str;
	profile_event[profile_idx].us_max = -1;
	profile_event[profile_idx].us_min = -1;
	profile_event[profile_idx].us_total = 0;
	profile_event[profile_idx].in_progress = 0;
	profile_event[profile_idx].new_data_to_print = 0;

	return profile_idx++;
}

void profile_pre_event(int id)
{
	if (id == -1)
		return;
#ifdef USE_GET_TIME_OF_DAY
	if (gettimeofday(&profile_event[id].tv, NULL)) {
		printf("%s:%d gettimeofday failed!", __FILE__, __LINE__);
		return;
	}
#endif
#ifdef USE_CLOCK_GET_TIME
	if (clock_gettime(CLOCK_MONOTONIC, &profile_event[id].ts)) {
		printf("%s:%d clock_gettime failed!", __FILE__, __LINE__);
		return;
	}
#endif
	profile_event[id].in_progress = 1;
}

#ifdef USE_CLOCK_GET_TIME
#define NSEC_PER_SEC 1e9
static void timespec_sub(struct timespec *lhs, struct timespec *rhs,
		struct timespec *res)
{
	res->tv_sec = lhs->tv_sec - rhs->tv_sec;
	res->tv_nsec = lhs->tv_nsec - rhs->tv_nsec;
	if (res->tv_nsec < 0) {
		res->tv_nsec += NSEC_PER_SEC;
		res->tv_sec--;
	}
}
#endif

void profile_post_event(int id)
{
	/* protection against bad id or event not in progress 
	 * (can occur if post is called before pre)
	 */
	if (id == -1 || !profile_event[id].in_progress)
		return;

	int us;
#ifdef USE_GET_TIME_OF_DAY
	{
		struct timeval tv, res;
		if (gettimeofday(&tv, NULL)) {
			printf("%s:%d gettimeofday failed!", __FILE__, __LINE__);
			return;
		}
		timersub(&tv, &profile_event[id].tv, &res);
		us = res.tv_usec + (res.tv_sec ? res.tv_sec * 1000000 : 0);
	}
#endif

#ifdef USE_CLOCK_GET_TIME
	{
		struct timespec ts, res;
		if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
			printf("%s:%d clock_gettime failed!", __FILE__, __LINE__);
			return;
		}
		timespec_sub(&ts, &profile_event[id].ts, &res);
		us = res.tv_nsec / 1000 + (res.tv_sec ? res.tv_sec * NSEC_PER_SEC : 0);
	}
#endif

	profile_event[id].us_last = us;

	/* Note that there's no overflow protection implemented here.
	 * Until implemented, the avaerage will be correct until overflow occures 
	 */
	profile_event[id].us_total += us;
	if (profile_event[id].us_max == -1 || profile_event[id].us_max < us)
		profile_event[id].us_max = us;
	if (profile_event[id].us_min == -1 || profile_event[id].us_min > us)
		profile_event[id].us_min = us;
	profile_event[id].num_hits++;
	profile_event[id].new_data_to_print = 1;
	profile_event[id].in_progress = 0;
}
