/* GStreamer
 *
 * Copyright (C) 2002 Ronald Bultje <rbultje@ronald.bitfreak.net>
 *               2006 Edgard Lima <edgard.lima@indt.org.br>
 *
 * v4l2_calls.c - generic V4L2 calls handling
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#ifdef __sun
/* Needed on older Solaris Nevada builds (72 at least) */
#include <stropts.h>
#include <sys/ioccom.h>
#endif
#include "v4l2_calls.h"
#include "gstv4l2src.h"

/******************************************************
 * gst_v4l2_get_capabilities():
 *   get the device's capturing capabilities
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2_get_capabilities(GstV4l2Object * v4l2object)
{
	GstElement *e;
	gboolean ret = FALSE;

	e = v4l2object->element;

	GST_DEBUG_OBJECT(e, "getting capabilities");

	if (!GST_V4L2_IS_OPEN(v4l2object))
		goto gst_v4l2_get_capabilities_exit;

	if (ioctl(v4l2object->video_fd, VIDIOC_QUERYCAP, &v4l2object->vcap) < 0)
		goto cap_failed;

	GST_LOG_OBJECT(e, "driver:      '%s'", v4l2object->vcap.driver);
	GST_LOG_OBJECT(e, "card:        '%s'", v4l2object->vcap.card);
	GST_LOG_OBJECT(e, "bus_info:    '%s'", v4l2object->vcap.bus_info);
	GST_LOG_OBJECT(e, "version:     %08x", v4l2object->vcap.version);
	GST_LOG_OBJECT(e, "capabilites: %08x", v4l2object->vcap.capabilities);

	ret = TRUE;
	goto gst_v4l2_get_capabilities_exit;

	/* ERRORS */
      cap_failed:
	{
		GST_ELEMENT_ERROR(v4l2object->element, RESOURCE, SETTINGS,
				  (_
				   ("Error getting capabilities for device '%s': "
				    "It isn't a v4l2 driver. Check if it is a v4l1 driver."),
				   v4l2object->videodev), GST_ERROR_SYSTEM);
		ret = FALSE;
	}
      gst_v4l2_get_capabilities_exit:
	return ret;
}

#define CLOSE_IF_OPEN(fd) \
	do { \
		if (fd != -1) \
			close(fd); \
		fd = -1; \
	} while (0)

/******************************************************
 * gst_v4l2_open():
 *   open the video device (v4l2object->videodev)
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2_open(GstV4l2Object * v4l2object)
{
	gboolean ret = FALSE;

	GST_DEBUG_OBJECT(v4l2object->element, "Trying to open device %s",
			 v4l2object->videodev);

	GST_V4L2_CHECK_NOT_OPEN(v4l2object);
	GST_V4L2_CHECK_NOT_ACTIVE(v4l2object);

	/* be sure we have a device */
	if (!v4l2object->videodev)
		v4l2object->videodev = g_strdup("/dev/video");

	PROF_SECT_DECL_BEGIN(power_on_sensor);
	/* open the device */
	if ((v4l2object->video_fd = open(v4l2object->videodev, O_RDWR)) == -1)
		goto not_open;
	if ((v4l2object->still_fd = open(v4l2object->videodev, O_RDWR)) == -1)
		goto not_open;

	GST_INFO_OBJECT(v4l2object->element, "video fd opened");
	PROF_SECT_END(power_on_sensor);

	if (pipe(v4l2object->propset_notify_fd))
		goto pipe_create_error;

	/* get capabilities, error will be posted */
	if (!gst_v4l2_get_capabilities(v4l2object))
		goto error;

	/* do we need to be a capture device? */
	if (GST_IS_V4L2SRC(v4l2object) &&
	    !(v4l2object->vcap.capabilities & V4L2_CAP_VIDEO_CAPTURE))
		goto not_capture;

	GST_INFO_OBJECT(v4l2object->element,
			"Opened device '%s' (%s) successfully",
			v4l2object->vcap.card, v4l2object->videodev);

	ret = TRUE;
	goto gst_v4l2_open_exit;

	/* ERRORS */
pipe_create_error:
	{
		GST_ELEMENT_ERROR(v4l2object->element, RESOURCE, OPEN_READ_WRITE,
				(_("pipe failed")), GST_ERROR_SYSTEM);
		goto error;
	}
not_open:
	{
		GST_ELEMENT_ERROR(v4l2object->element, RESOURCE,
				OPEN_READ_WRITE,
				(_("Could not open device '%s' for reading and writing."),
				 v4l2object->videodev), GST_ERROR_SYSTEM);
		goto error;
	}
not_capture:
	{
		GST_ELEMENT_ERROR(v4l2object->element, RESOURCE, NOT_FOUND,
				(_("Device '%s' is not a capture device."),
				 v4l2object->videodev),
				("Capabilities: 0x%x",
				 v4l2object->vcap.capabilities));
		goto error;
	}
error:
	{
		CLOSE_IF_OPEN(v4l2object->video_fd);
		CLOSE_IF_OPEN(v4l2object->still_fd);
		CLOSE_IF_OPEN(v4l2object->propset_notify_fd[0]);
		CLOSE_IF_OPEN(v4l2object->propset_notify_fd[1]);

		ret = FALSE;
	}
gst_v4l2_open_exit:
	return ret;
}

/******************************************************
 * gst_v4l2_close():
 *   close the video device (v4l2object->video_fd)
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2_close(GstV4l2Object * v4l2object)
{
	GST_INFO_OBJECT(v4l2object->element, "closing %s",
			 v4l2object->videodev);

	GST_V4L2_CHECK_OPEN(v4l2object);
	GST_V4L2_CHECK_NOT_ACTIVE(v4l2object);

	/* close device */
	CLOSE_IF_OPEN(v4l2object->video_fd);
	CLOSE_IF_OPEN(v4l2object->still_fd);
	CLOSE_IF_OPEN(v4l2object->propset_notify_fd[0]);
	CLOSE_IF_OPEN(v4l2object->propset_notify_fd[1]);

	return TRUE;
}

/******************************************************
 * gst_v4l2_get_attribute():
 *   try to get the value of one specific attribute
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean
gst_v4l2_get_attribute(GstV4l2Object * v4l2object,
		       int attribute_num, int *value)
{
	struct v4l2_control control;

	GST_DEBUG_OBJECT(v4l2object->element, "getting value of attribute %d",
			 attribute_num);

	if (!GST_V4L2_IS_OPEN(v4l2object))
		return FALSE;

	control.id = attribute_num;

	if (ioctl(v4l2object->video_fd, VIDIOC_G_CTRL, &control) < 0)
		goto ctrl_failed;

	*value = control.value;

	return TRUE;

	/* ERRORS */
      ctrl_failed:
	{
		GST_ELEMENT_WARNING(v4l2object->element, RESOURCE, SETTINGS,
				    (_
				     ("Failed to get value for control %d on device '%s'."),
				     attribute_num, v4l2object->videodev),
				    GST_ERROR_SYSTEM);
		return FALSE;
	}
}

/******************************************************
 * gst_v4l2_set_attribute():
 *   try to set the value of one specific attribute
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean
gst_v4l2_set_attribute(GstV4l2Object * v4l2object,
		       int attribute_num, const int value)
{
	struct v4l2_control control;

	GST_DEBUG_OBJECT(v4l2object->element,
			 "setting value of attribute %d to %d", attribute_num,
			 value);

	if (!GST_V4L2_IS_OPEN(v4l2object))
		return FALSE;

	control.id = attribute_num;
	control.value = value;
	if (ioctl(v4l2object->video_fd, VIDIOC_S_CTRL, &control) < 0)
		goto ctrl_failed;

	return TRUE;

	/* ERRORS */
      ctrl_failed:
	{
		GST_ELEMENT_WARNING(v4l2object->element, RESOURCE, SETTINGS,
				    (_
				     ("Failed to set value %d for control %d on device '%s'."),
				     value, attribute_num,
				     v4l2object->videodev), GST_ERROR_SYSTEM);
		return FALSE;
	}
}

gboolean gst_v4l2_get_input(GstV4l2Object * v4l2object, gint * input)
{
	gint n;

	GST_DEBUG_OBJECT(v4l2object->element, "trying to get input");

	if (!GST_V4L2_IS_OPEN(v4l2object))
		return FALSE;

	if (ioctl(v4l2object->video_fd, VIDIOC_G_INPUT, &n) < 0)
		goto input_failed;

	*input = n;

	GST_DEBUG_OBJECT(v4l2object->element, "input: %d", n);

	return TRUE;

	/* ERRORS */
      input_failed:
	{
		GST_ELEMENT_WARNING(v4l2object->element, RESOURCE, SETTINGS,
				    (_
				     ("Failed to get current input on device '%s'. May be it is a radio device"),
				     v4l2object->videodev), GST_ERROR_SYSTEM);
		return FALSE;
	}
}

gboolean gst_v4l2_set_input(GstV4l2Object * v4l2object, gint input)
{
	GST_DEBUG_OBJECT(v4l2object->element, "trying to set input to %d",
			 input);

	if (!GST_V4L2_IS_OPEN(v4l2object))
		return FALSE;

	if (ioctl(v4l2object->video_fd, VIDIOC_S_INPUT, &input) < 0)
		goto input_failed;

	return TRUE;

	/* ERRORS */
      input_failed:
	{
		GST_ELEMENT_WARNING(v4l2object->element, RESOURCE, SETTINGS,
				    (_("Failed to set input %d on device %s."),
				     input, v4l2object->videodev),
				    GST_ERROR_SYSTEM);
		return FALSE;
	}
}
